-- prepared by Aaron Buma
-- http://AaronBuma.com
-- twitter:  @AaronDBuma

RAISERROR ('Dont run the whole script, just run sections at a time', 20, 1)  WITH LOG;

------------------------ EXAMPLES SETUP ------------------------

USE Master;
GO

IF EXISTS(SELECT name 
            FROM sys.databases 
		   WHERE name = 'QueryTraining')
BEGIN
	ALTER DATABASE [QueryTraining] 
	  SET SINGLE_USER WITH ROLLBACK IMMEDIATE;
	DROP DATABASE [QueryTraining];
END
CREATE DATABASE [QueryTraining];
GO

USE [QueryTraining];


CREATE TABLE [dbo].[Inventory](
	[InventoryID] [INTEGER] IDENTITY(-2147483647,1) PRIMARY KEY,
	[CategoryID] [INT], 
	[ItemSKU] [VARCHAR](200),
	[Description] [VARCHAR](4000),
	[QuantityOnHand] [INT],
	[UnitCost] DECIMAL(10,2),
	[UnitRetail] DECIMAL(10,2),
	[Archived] BIT
); 


CREATE TABLE [dbo].[Categories](
	[CategoryID] [INTEGER],
	[CategoryTitle] [VARCHAR](200),
	[Description] [VARCHAR](200),
	[Archived] BIT
); 

INSERT INTO dbo.Categories ( CategoryID, CategoryTitle, Description, Archived )
VALUES  
(1, 'SLEEPINGBAG','Body Socks', 0),
(2, 'TENT','Suspended Tarps', 0),
(3, 'AXE','Finger Removers', 0),
(4, 'FOOD','Animal Bait', 0);

INSERT INTO dbo.Inventory (CategoryID, ItemSKU, Description, QuantityOnHand, UnitCost, UnitRetail, Archived)
VALUES
(1, 'SB001-LG-RED', 'Sleeping Bag - Large - RED', 5, 5.75, 19.99, 0),
(1, 'SB001-LG-GRN', 'Sleeping Bag - Large - GREEN', 13, 5.75, 14.99, 1),
(1, 'SB001-LG-BLK', 'Sleeping Bag - Large - BLACK', 1, 5.75, 19.99, 0),
(2, 'TT001-10-CMO', 'Tent - 10 Man - CAMO', 5, 10.00, 99.95, 0),
(3, 'AX013-HT-BLK', 'Axe - Hatchet - Black', 35, 4.99, 24.95, 0),
(4, 'MR005-1S-SPG', 'MRE - 1 Serve - Spaghetti', 5, 0.50, 7.99, 0),
(4, 'MR006-1A-PIZ', 'L3i987??113', 1, NULL, 99.95, 0);
--------------------------------------------------------------------------
--Basic Subqueries, or Derived Tables

--Our results without a subquery, but requires knowledge of the Category ID's
SELECT *
  FROM dbo.Inventory AS i
 WHERE i.CategoryID IN(1,2);
 
--Our results without a subquery, only knowing the Category Title
SELECT *
  FROM dbo.Inventory AS i
 WHERE i.CategoryID IN(
			
			SELECT c.CategoryID
			  FROM dbo.Categories AS c
			 WHERE c.CategoryTitle IN('SLEEPINGBAG','TENT')
			 
			 );
			 
--Correlated Subqueries
--How to get a count of how many items exist per category
SELECT *
  FROM dbo.Inventory AS i;
  
SELECT *
  FROM dbo.Categories AS c;
  
SELECT i.CategoryID, --Column to reference
	   COUNT(i.InventoryID) AS ItemsInCategory --Values we want
  FROM dbo.Inventory AS i
 GROUP BY i.CategoryID;

--How can we get the same data into results from the category table?
SELECT *,
	   (
	   SELECT COUNT(i2.InventoryID)
		 FROM dbo.Inventory AS i2
		WHERE i2.CategoryID = c.categoryID --This is the correlation
	   ) AS ItemsInCategory
  FROM dbo.Categories AS c;
 

--How to list items along with a count of how many items are in the the same category
SELECT *,
	   ( 
		SELECT COUNT(i2.InventoryID) AS ItemsInCategory --Values we want
		  FROM dbo.Inventory AS i2
		 WHERE i2.CategoryID = i.CategoryID --Column to reference
		 GROUP BY i2.CategoryID
	   ) AS ItemsInCategory
  FROM dbo.Inventory AS i;
  
--With that same query, how do we exclude the current item?
SELECT *,
	   ( 
		SELECT COUNT(i2.InventoryID) AS ItemsInCategory --Values we want
		  FROM dbo.Inventory AS i2
		 WHERE i2.CategoryID = i.CategoryID --Column to reference
		   AND i2.InventoryID != i.inventoryID --Column to reference
		 GROUP BY i2.CategoryID
	   ) AS OtherItemsInCategory
  FROM dbo.Inventory AS i;
  
--Add ISNULL to account for rows where there is no other items
SELECT *,
	   ISNULL((
		SELECT COUNT(i2.InventoryID) AS ItemsInCategory --Values we want
		  FROM dbo.Inventory AS i2
		 WHERE i2.CategoryID = i.CategoryID --Column to reference
		   AND i2.InventoryID ! = i.inventoryID --Column to reference
		 GROUP BY i2.CategoryID
	   ),0) AS OtherItemsInCategory
  FROM dbo.Inventory AS i;

---------------------------------------------------------------------------------------------------------  
--SubQuery Extensions: ALL, ANY, SOME and EXISTS

--Return a list of categories, that have with quantity on hand greater than or equal to 5
SELECT c.CategoryTitle, i.*
  FROM dbo.Inventory AS i
 INNER JOIN dbo.Categories AS c ON c.CategoryID = i.CategoryID;
   
SELECT *
  FROM dbo.Categories AS c
 WHERE 5 <= ALL (
	SELECT i.QuantityOnHand
      FROM dbo.Inventory AS i 
     WHERE i.CategoryID = c.CategoryID
	);
--ALL Operator:
--CategoryID 1 has 1 record  where the QuantityOnHand is below 5, ALL expression is FALSE
--CategoryID 2 has 0 records where the QuantityOnHand is below 5, ALL expression is TRUE - It is in the results 
--CategoryID 3 has 1 record  where the QuantityOnHand is below 5, ALL expression is TRUE - It is in the results
--CategoryID 4 has 1 record  where the QuantityOnHand is below 5, ALL expression is FALSE

SELECT c.CategoryTitle, i.*
  FROM dbo.Inventory AS i
 INNER JOIN dbo.Categories AS c ON c.CategoryID = i.CategoryID;
  
SELECT *
  FROM dbo.Categories AS c
 WHERE 5 <= ANY (
	SELECT i.QuantityOnHand
      FROM dbo.Inventory AS i 
     WHERE i.CategoryID = c.CategoryID
	);
--ANY Operator:
--CategoryID 1 has 2 records where the QuantityOnHand is ABOVE or EQUAL TO 5, ANY expression is TRUE - It is in the results 
--CategoryID 2 has 1 record  where the QuantityOnHand is ABOVE or EQUAL TO 5, ANY expression is TRUE - It is in the results 
--CategoryID 3 has 1 record  where the QuantityOnHand is ABOVE or EQUAL TO 5, ANY expression is TRUE - It is in the results
--CategoryID 4 has 1 record  where the QuantityOnHand is ABOVE or EQUAL TO 5, ANY expression is TRUE - It is in the results 
 
--SOME Operator is the ISO Equivalent of ANY
SELECT *
  FROM dbo.Categories AS c
 WHERE 5 <= SOME (
	SELECT i.QuantityOnHand
      FROM dbo.Inventory AS i 
     WHERE i.CategoryID = c.CategoryID
	);
	
	
--EXISTS
SELECT *
  FROM dbo.Categories AS c
 WHERE EXISTS (
	SELECT 'ThisIsNotReturned'
      FROM dbo.Inventory AS i 
     WHERE i.CategoryID = c.CategoryID
       AND i.QuantityOnHand >= 5
	);
	
--It is the equivalent to 'SOME', it returns 'true' when first match is found.
--It is faster than ANY and SOME becuase it stops processing the subquery when the first match is found.